#!/usr/bin/python -tt
#
# - Cameron Simpson <cs@cskk.id.au>
#

''' Utility routines to parse data streams into Blocks and Block streams into IndirectBlocks.
'''

from heapq import heappush, heappop
from itertools import chain
import sys
from cs.buffer import CornuCopyBuffer
from cs.logutils import warning, exception
from cs.pfx import Pfx, PfxThread
from cs.queues import IterableQueue
from cs.seq import tee
from .block import Block, IndirectBlock
from .scan import scanbuf

# constraints on the chunk sizes yields from blocked_chunks_of
MIN_BLOCKSIZE = 80          # less than this seems silly
MAX_BLOCKSIZE = 16383       # fits in 2 octets BS-encoded

# default read size for file scans
DEFAULT_SCAN_SIZE = 1024 * 1024

def top_block_for(blocks):
  ''' Return a top Block for a stream of Blocks.
  '''
  # obtain stream of full indirect blocks from `blocks`
  blocks = indirect_blocks(blocks)

  # Fetch the first two indirect blocks from the generator.
  # If there is none, return a single empty direct block.
  # If there is just one, return it directly.
  # Otherwise there are at least two:
  # replace the blocks with another level of indirect_blocks()
  # reading from the two fetched blocks and the tail of the current
  # blocks then lather, rinse, repeat.
  #
  while True:
    try:
      topblock = next(blocks)
    except StopIteration:
      # no blocks - return the empty block - no data
      return Block(data=b'')

    # we have a full IndirectBlock
    # if there are more, replace our blocks with
    #   indirect_blocks(topblock + nexttopblock + blocks)
    try:
      nexttopblock = next(blocks)
    except StopIteration:
      # just one IndirectBlock - we're done
      return topblock

    # add a layer of indirection and repeat
    blocks = indirect_blocks(chain( ( topblock, nexttopblock ), blocks ))

  raise RuntimeError("SHOULD NEVER BE REACHED")

def indirect_blocks(blocks):
  ''' A generator that yields full IndirectBlocks from an iterable
      source of Blocks, except for the last Block which need not
      necessarily be bundled into an IndirectBlock.
  '''
  subblocks = []
  subsize = 0
  for block in blocks:
    enc = block.encode()
    if subsize + len(enc) > MAX_BLOCKSIZE:
      # overflow
      if not subblocks:
        # do not yield empty indirect block, flag logic error instead
        warning(
            "no pending subblocks at flush, presumably len(block.encode()) %d > MAX_BLOCKSIZE %d",
            len(enc), MAX_BLOCKSIZE)
      else:
        yield IndirectBlock(subblocks)
        subblocks = []
        subsize = 0
    subblocks.append(block)
    subsize += len(enc)

  # handle the termination case
  if subblocks:
    if len(subblocks) == 1:
      # one block unyielded - don't bother wrapping into an iblock
      block = subblocks[0]
    else:
      block = IndirectBlock(subblocks)
    yield block

def blockify(chunks, scanner=None, min_block=None, max_block=None):
  ''' Wrapper for blocked_chunks_of which yields Blocks from the data chunks.
  '''
  for chunk in blocked_chunks_of(chunks, scanner, min_block=min_block, max_block=max_block):
    yield Block(data=chunk)

def spliced_blocks(B, new_blocks):
  ''' Splice an iterable of (offset, Block) into the data of the Block `B`.
      Yield high level blocks.
  '''
  upto = 0      # data span yielded so far
  for offset, newB in new_blocks:
    if offset > upto:
      yield from B.top_blocks(upto, offset)
      upto = offset
    yield newB
    upto += len(newB)
  if upto < len(B):
    yield from B.top_blocks(upto, len(B))

class _PendingBuffer:
  ''' Class to manage the unbound chunks accrued by blocked_chunks_of below.
  '''

  def __init__(self, max_block, offset=0):
    if max_block < 1:
      raise ValueError("max_block must be >= 1, received: %s" % (max_block,))
    self.max_block = max_block
    self.pending = []
    self.pending_room = self.max_block
    self.offset = offset

  def flush(self):
    ''' Yield the pending chunks joined together, if any.
        Advance the offset.
    '''
    if self.pending:
      assert self.pending_room < self.max_block
      chunk = b''.join(self.pending)
      self.pending = []
      self.pending_room = self.max_block
      self.offset += len(chunk)
      yield chunk

  def append(self, chunk):
    ''' Append `chunk` to the pending buffer.
        Yield any overflow.
    '''
    pending_room = self.pending_room
    while len(chunk) >= pending_room:
      self.pending.append(chunk[:pending_room])
      self.pending_room = 0
      yield from self.flush()
      chunk = chunk[pending_room:]
      pending_room = self.pending_room
    if chunk:
      self.pending.append(chunk)
      self.pending_room -= len(chunk)

def blocked_chunks_of(
    chunks,
    scanner=None,
    min_block=None, max_block=None,
    histogram=None,
):
  ''' Generator which connects to a scanner of a chunk stream in
      order to emit low level edge aligned data chunks.

      Parameters:
      * `chunks`: a source iterable of data chunks, handed to `scanner`
      * `scanner`: optional callable accepting a CornuCopyBuffer and
        returning an iterable of ints, such as a generator. `scanner`
        may be None, in which case only the rolling hash is used
        to locate boundaries.
      * `min_block`: the smallest amount of data that will be used
        to create a Block, default MIN_BLOCKSIZE
      * `max_block`: the largest amount of data that will be used to
        create a Block, default MAX_BLOCKSIZE
      * `histogram`: if not None, a defaultdict(int) to collate counts.
        Integer indices count block sizes and string indices are used
        for 'bytes_total' and 'bytes_hash_scanned'.

      The iterable returned from `scanner(chunks)` yields ints which are
      considered desirable block boundaries.
  '''
  # pylint: disable=too-many-nested-blocks,too-many-statements
  # pylint: disable=too-many-branches,too-many-locals
  with Pfx("blocked_chunks_of"):
    if min_block is None:
      min_block = MIN_BLOCKSIZE
    elif min_block < 8:
      raise ValueError("rejecting min_block < 8: %s" % (min_block,))
    if max_block is None:
      max_block = MAX_BLOCKSIZE
    elif max_block >= 1024*1024:
      raise ValueError("rejecting max_block >= 1024*1024: %s" % (max_block,))
    if min_block >= max_block:
      raise ValueError("rejecting min_block:%d >= max_block:%d"
                       % (min_block, max_block))
    # obtain iterator of chunks; this avoids accidentally reusing the chunks
    # if for example chunks is a sequence
    chunk_iter = iter(chunks)
    # Set up parseQ, an iterable yielding a mix of source data and
    # offsets representing desirable block boundaries.
    # If there is no scanner, this is just chunk_iter.
    # If there is a scanner we dispatch the scanner in a separate
    # Thread and feed it a tee() of chunk_iter, which copies chunks
    # to the parseQ when chunks are obtained by the scanner. The
    # Thread runs the scanner and copies its output offsets to the
    # parseQ.
    # The tee() arranges that chunks arrive before any offsets within them.
    if scanner is None:
      # No scanner, consume the chunks directly.
      parseQ = chunk_iter
    else:
      # Consume the chunks and offsets via a queue.
      # The scanner puts offsets onto the queue.
      # When the scanner fetches from the chunks, those chunks are copied to the queue.
      # When the scanner terminates, any remaining chunks are also copied to the queue.
      parseQ = IterableQueue()
      chunk_iter = tee(chunk_iter, parseQ)
      def run_parser():
        ''' Thread body to run the supplied scanner against the input data.
        '''
        bfr = CornuCopyBuffer(chunk_iter)
        # pylint: disable=broad-except
        try:
          for offset in scanner(bfr):
            # the scanner should yield only offsets, not chunks and offsets
            if not isinstance(offset, int):
              warning("discarding non-int from scanner %s: %s", scanner, offset)
            else:
              parseQ.put(offset)
        except Exception as e:
          exception("exception from scanner %s: %s", scanner, e)
        # Consume the remainder of chunk_iter; the tee() will copy it to parseQ.
        for _ in chunk_iter:
          pass
        # end of offsets and chunks
        parseQ.close()
      PfxThread(target=run_parser).run()
    # inbound chunks and offsets
    in_offsets = []     # heap of unprocessed edge offsets
    # prime `available_chunk` with the first data chunk, ready for get_next_chunk
    try:
      available_chunk = next(parseQ)
    except StopIteration:
      # no data! just return
      return
    def get_next_chunk():
      ''' Fetch and return the next data chunk from the `parseQ`.
          Return None at end of input.
          Also gather all the following offsets from the queue before return.
          Because this inherently means collecting the chunk beyond
          these offsets, we keep that in `available_chunk` for the
          next call.
          Sets parseQ to None if the end of the iterable is reached.
      '''
      nonlocal parseQ, in_offsets, hash_value, available_chunk
      if parseQ is None:
        assert available_chunk is None
        return None
      next_chunk = available_chunk
      available_chunk = None
      assert not isinstance(next_chunk, int)
      # scan the new chunk and load potential edges into the offset heap
      hash_value, chunk_scan_offsets = scanbuf(hash_value, next_chunk)
      for cso in chunk_scan_offsets:
        heappush(in_offsets, offset + cso)
      # gather items from the parseQ until the following chunk
      # or end of input
      while True:
        try:
          item = next(parseQ)
        except StopIteration:
          parseQ = None
          break
        else:
          if isinstance(item, int):
            heappush(in_offsets, item)
          else:
            available_chunk = item
            break
      return next_chunk
    last_offset = None
    first_possible_point = None
    max_possible_point = None
    def recompute_offsets():
      ''' Recompute relevant offsets from the block parameters.
          The first_possible_point is last_offset+min_block,
            the earliest point at which we will accept a block boundary.
          The max_possible_point is last_offset+max_block,
            the latest point at which we will accept a block boundary;
            we will choose this if no next_offset or hash offset
            is found earlier.
      '''
      nonlocal last_offset, first_possible_point, max_possible_point
      first_possible_point = last_offset + min_block
      max_possible_point   = last_offset + max_block
      ##X("recomputed offsets: last_offset=%d, first_possible_point=%d, max_possible_point=%d",
      ##  last_offset, first_possible_point, max_possible_point)
    # prepare initial state
    last_offset = 0         # latest released boundary
    recompute_offsets()     # compute first_possible_point and max_possible_point
    hash_value = 0
    offset = 0
    chunk0 = None
    offset0 = None
    # unblocked outbound data
    pending = _PendingBuffer(max_block)
    # Read data chunks and locate desired boundaries.
    while True:
      chunk = get_next_chunk()
      if chunk is None:
        break
      # verify current chunk start offset against end of previous chunk
      assert chunk0 is None or offset == offset0 + len(chunk0), \
          "offset0=%d, len(chunk0)=%d: sum(%d) != current offset %d" \
          % (offset0, len(chunk0), offset0 + len(chunk0), offset)
      chunk0 = chunk
      offset0 = offset
      chunk = memoryview(chunk)
      chunk_end_offset = offset + len(chunk)
      # process current chunk
      advance_by = 0    # how much data to add to the pending buffer
      release = False   # whether we hit a boundary ==> flush the buffer
      while chunk:
        if advance_by > 0:
          # advance through this chunk
          # buffer the advance
          # release ==> flush the buffer and update last_offset
          assert advance_by is not None
          assert advance_by >= 0
          assert advance_by <= len(chunk)
          # save the advance bytes and yield any overflow
          for out_chunk in pending.append(chunk[:advance_by]):
            yield out_chunk
            if histogram is not None:
              out_chunk_size = len(out_chunk)
              histogram['bytes_total'] += out_chunk_size
              histogram[out_chunk_size] += 1
              histogram['buffer_overflow_chunks'] += 1
          offset += advance_by
          chunk = chunk[advance_by:]
          if last_offset != pending.offset:
            # if the flush discarded a full buffer we need to adjust our boundaries
            last_offset = pending.offset
            recompute_offsets()
          if release:
            release = False   # becomes true if we should flush after taking data
            # yield the current pending data
            for out_chunk in pending.flush():
              yield out_chunk
              if histogram is not None:
                out_chunk_size = len(out_chunk)
                histogram['bytes_total'] += out_chunk_size
                histogram[out_chunk_size] += 1
            last_offset = pending.offset
            recompute_offsets()
          if not chunk:
            # consumed the end of the chunk, need a new one
            break
        advance_by = None
        # fetch the next available edge, None if nothing available or suitable
        while True:
          if in_offsets:
            next_offset = heappop(in_offsets)
            if next_offset > offset and next_offset >= first_possible_point:
              break
          else:
            next_offset = None
            break
        if next_offset is None or next_offset > chunk_end_offset:
          # no suitable edge: consume the chunk and advance
          take_to = chunk_end_offset
        else:
          # edge before end of chunk: use it
          take_to = next_offset
          release = True
        advance_by = take_to - offset
        assert advance_by > 0
    # yield any left over data
    for out_chunk in pending.flush():
      yield out_chunk
      if histogram is not None:
        out_chunk_size = len(out_chunk)
        histogram['bytes_total'] += out_chunk_size
        histogram[out_chunk_size] += 1

if __name__ == '__main__':
  from .blockify_tests import selftest
  selftest(sys.argv)



#!/usr/bin/python
#
# The generic Store factory and parser for Store specifications.
#   - Cameron Simpson <cs@cskk.id.au> 20dec2016
#

''' Composition of Stores from text specifications.
'''

from os.path import isdir
from subprocess import Popen, PIPE
from cs.lex import skipwhite, get_identifier, get_qstr
from cs.pfx import Pfx
from .convert import get_integer
from .stream import StreamStore

def parse_store_specs(s, offset=0):
  ''' Parse the string `s` for a list of Store specifications.
  '''
  with Pfx("parse_store_specs(%r)", s):
    store_specs = []
    while offset < len(s):
      with Pfx("offset %d", offset):
        store_text, store_type, params, offset = get_store_spec(s, offset)
        store_specs.append( (store_text, store_type, params) )
      if offset < len(s):
        with Pfx("offset %d", offset):
          sep = s[offset]
          offset += 1
          if sep == ',':
            continue
          raise ValueError(
              "expected comma ',', found unexpected separator: %r"
              % (sep,))
    return store_specs

def get_store_spec(s, offset=0):
  ''' Get a single Store specification from a string.
      Return `(matched, type, params, offset)`
      being the matched text, store type, parameters and the new offset.

      Recognised specifications:
      * `"text"`: Quoted store spec, needed to enclose some of the following
        syntaxes if they do not consume the whole string.
      * `[clause_name]`: The name of a clause to be obtained from a Config.
      * `/path/to/something`, `./path/to/something`:
        A filesystem path to a local resource.
        Supported paths:
        - `.../foo.sock`: A UNIX socket based StreamStore.
        - `.../dir`: A DataDirStore directory.
        - `.../foo.vtd `: (STILL TODO): A DataFileStore.
      * `|command`: A subprocess implementing the streaming protocol.
      * `store_type(param=value,...)`:
        A general Store specification.
      * `store_type:params...`:
        An inline Store specification.
        Supported inline types: `tcp:[host]:port`

      TODO:
      * `ssh://host/[store-designator-as-above]`:
      * `unix:/path/to/socket`:
        Connect to a daemon implementing the streaming protocol.
      * `http[s]://host/prefix`:
        A Store presenting content under prefix:
        + `/h/hashcode.hashtype`: Block data by hashcode
        + `/i/hashcode.hashtype`: Indirect block by hashcode.
      * `s3://bucketname/prefix/hashcode.hashtype`:
        An AWS S3 bucket with raw blocks.
  '''
  offset0 = offset
  if offset >= len(s):
    raise ValueError("empty string")
  if s.startswith('"', offset):
    # "store_spec"
    qs, offset = get_qstr(s, offset, q='"')
    _, store_type, params, offset2 = get_store_spec(qs, 0)
    if offset2 < len(qs):
      raise ValueError("unparsed text inside quotes: %r" % (qs[offset2:],))
  elif s.startswith('[', offset):
    # [clause_name]
    store_type = 'config'
    offset = skipwhite(s, offset + 1)
    clause_name, offset = get_qstr_or_identifier(s, offset)
    offset = skipwhite(s, offset)
    if offset >= len(s) or s[offset] != ']':
      raise ValueError("offset %d: missing closing ']'" % (offset,))
    offset += 1
    params = {'clause_name': clause_name}
  elif s.startswith('/', offset) or s.startswith('./', offset):
    path = s[offset:]
    offset = len(s)
    if path.endswith('.sock'):
      store_type = 'socket'
      params = {'socket_path': path}
    elif isdir(path):
      store_type = 'datadir'
      params = {'path': path}
    else:
      raise ValueError(
          "%r: not a directory or a socket"
          % (path,))
  elif s.startswith('|', offset):
    # |shell command
    store_type = 'shell'
    params = {'shcmd': s[offset + 1:].strip()}
    offset = len(s)
  else:
    store_type, offset = get_identifier(s, offset)
    if not store_type:
      raise ValueError(
          "expected identifier at offset %d, found: %r"
          % (offset, s[offset:]))
    with Pfx(store_type):
      if s.startswith('(', offset):
        params, offset = get_params(s, offset + 1, ')')
      elif s.startswith(':', offset):
        offset += 1
        params = {}
        if store_type == 'tcp':
          colon2 = s.find(':', offset)
          if colon2 < offset:
            raise ValueError("missing second colon after offset %d" % (offset,))
          hostpart = s[offset:colon2]
          offset = colon2 + 1
          if not isinstance(hostpart, str):
            raise ValueError(
                "expected hostpart to be a string, got: %r" % (hostpart,))
          if not hostpart:
            hostpart = 'localhost'
          params['host'] = hostpart
          portpart, offset = get_token(s, offset)
          params['port'] = portpart
        else:
          raise ValueError("unrecognised Store type for inline form")
      else:
        raise ValueError("no parameters")
  return s[offset0:offset], store_type, params, offset

def get_params(s, offset, endchar):
  ''' Parse "param=value,...)". Return params dict and new offset.
  '''
  params = {}
  first = True
  while True:
    ch = s[offset:offset + 1]
    if not ch:
      raise ValueError("hit end of string, expected param or %r" % (endchar,))
    if ch == endchar:
      offset += 1
      return params, offset
    if not first and ch == ',':
      offset += 1
    param, offset = get_qstr_or_identifier(s, offset)
    if not param:
      raise ValueError("rejecting empty parameter name at: %r" % (s[offset:],))
    offset = skipwhite(s, offset)
    ch = s[offset:offset + 1]
    if not ch:
      raise ValueError("hit end of string after param %r, expected '='" % (param,))
    if ch != '=':
      raise ValueError("expected '=', found %r" % (ch,))
    offset = skipwhite(s, offset + 1)
    ch = s[offset:offset + 1]
    if not ch:
      raise ValueError("hit end of string after param %r=, expected token" % (param,))
    if ch == endchar or ch == ',':
      raise ValueError("expected token, found %r" % (ch,))
    value, offset = get_token(s, offset)
    params[param] = value
    first = False

def get_qstr_or_identifier(s, offset):
  ''' Parse q quoted string or an identifier.
  '''
  if s.startswith('"', offset):
    return get_qstr(s, offset, q='"')
  return get_identifier(s, offset)

def get_token(s, offset):
  ''' Parse an integer value, an identifier or a quoted string.
  '''
  if offset == len(s):
    raise ValueError("unexpected end of string, expected token")
  if s[offset].isdigit():
    token, offset = get_integer(s, offset)
  else:
    token, offset = get_qstr_or_identifier(s, offset)
  return token, offset

def CommandStore(shcmd, addif=False):
  ''' Factory to return a StreamStore talking to a command.
  '''
  name = "StreamStore(%r)" % ("|" + shcmd, )
  P = Popen(shcmd, shell=True, stdin=PIPE, stdout=PIPE)
  return StreamStore(name, P.stdin, P.stdout, local_store=None, addif=addif)



#!/usr/bin/python

''' Directory entry metadata.
'''

from __future__ import print_function
from functools import lru_cache
import json
import os
from collections import namedtuple
from pwd import getpwuid, getpwnam
from grp import getgrgid, getgrnam
from stat import S_ISUID, S_ISGID
import time
from cs.logutils import error, warning
from cs.serialise import get_bss, get_bsdata
from cs.threads import locked
from cs.x import X
from . import RLock
from .transcribe import Transcriber, register as register_transcriber, \
                        transcribe_mapping, parse_mapping

DEFAULT_DIR_ACL = 'o:rwx-'
DEFAULT_FILE_ACL = 'o:rw-x'

NOUSERID = -1
NOGROUPID = -1

@lru_cache(maxsize=64)
def username(uid):
  ''' Look up the login name associated with the supplied `uid`.
      Return None if unknown. Caches results, including lookup failure.
  '''
  try:
    pw = getpwuid(uid)
  except KeyError:
    return None
  return pw.pw_name

@lru_cache(maxsize=64)
def userid(username):
  ''' Look up the user id associated with the supplied `username`.
      Return None if unknown. Caches results, including lookup failure.
  '''
  try:
    pw = getpwnam(username)
  except KeyError:
    return None
  return pw.pw_uid

@lru_cache(maxsize=64)
def groupname(gid):
  ''' Look up the group name associated with the supplied `gid`.
      Return None if unknown. Caches results, including lookup failure.
  '''
  try:
    gr = getgrgid(gid)
  except KeyError:
    return None
  return gr.gr_name

@lru_cache(maxsize=64)
def groupid(groupname):
  ''' Look up the group id associated with the supplied `groupname`.
      Return None if unknown. Caches results, including lookup failure.
  '''
  try:
    gr = getgrnam(groupname)
  except KeyError:
    return None
  return gr.gr_gid

def permbits_to_allow_deny(bits):
  ''' Take a UNIX 3-bit permission value and return the ACL allow and deny strings.
      Example: 6 (110) => 'rw', 'x'
  '''
  add = ''
  sub = ''
  for c, bit in ('r', 0x04), ('w', 0x02), ('x', 0x01):
    if bits&bit:
      add += c
    else:
      sub += c
  return add, sub

def rwx(mode):
  ''' Transcribe 3 bits of a UNIX mode in 'rwx' form.
  '''
  return (
      ( 'r' if mode&4 else '-' )
      + ( 'w' if mode&2 else '-' )
      + ( 'x' if mode&1 else '-' )
  )

class AC(namedtuple('AccessControl', 'audience allow deny')):
  ''' An Access Control.
  '''

  def __str__(self):
    ''' Encode this access control as text.
    '''
    audience, allow, deny = self
    if ':' in audience:
      raise ValueError(
          "invalid audience, may not contain a colon: %r"
          % (audience,))
    if '-' in allow:
      raise ValueError(
          "invalid allow, may not contain a dash: %r"
          % (allow,))
    return audience + ':' + allow + '-' + deny

  def __repr__(self):
    return ':'.join( (type(self).__name__, str(self)) )

  @classmethod
  def from_str(cls, ac_text):
    ''' Factory function to return a new AC from an encoded AC.
    '''
    audience, allow_deny = ac_text.split(':', 1)
    if audience not in ('o', 'g', '*'):
      raise ValueError("invalid audience %r from %r" % (audience, ac_text))
    allow, deny = allow_deny.split('-', 1)
    return cls(audience, allow, deny)

  def __call__(self, M, accesses):
    ''' Call the AC with the Meta `M` and the required permissions `accesses`.

        *Note*: this does _not_ check the audience and presumes
        that the access control is applicable.

        Returns False if any access is in .deny, otherwise returns True if
        all accesses are in .allow.

        Subclasses override this method and check the access control
        audience for applicability before deferring to this method
        to test access; they return None if the access control is
        not applicable, for example an AC_Owner when the Meta has
        no owner or the Meta owner does not match the caller owner.
    '''
    allow = self.allow
    deny = self.deny
    for access in accesses:
      if access in deny:
        return False
      if access not in allow:
        return False
    return True

  @property
  def unixmode(self):
    ''' Return the 3-bit UNIX mode for this access.
    '''
    mode = 0
    for a in self.allow:
      if a == 'r':
        mode |= 4
      elif a == 'w':
        mode |= 2
      elif a == 'x':
        mode |= 1
      else:
        warning("AC.unixmode: ignoring unsupported permission %r", a)
    return mode

AC_Owner = lambda allow, deny: AC('o', allow, deny)
AC_Group = lambda allow, deny: AC('g', allow, deny)
AC_Other = lambda allow, deny: AC('*', allow, deny)

class ACL(list):
  ''' An access control list.
  '''

  def __str__(self):
    ''' Transcribe a list of AC instances as text.
    '''
    return ','.join( [ str(ac) for ac in self ] )

  @classmethod
  def from_str(cls, acl_text):
    ''' Return an ACL from the str `acl_text`.
    '''
    acl = cls()
    for ac_text in acl_text.split(','):
      if ac_text:
        try:
          ac = AC.from_str(ac_text)
        except ValueError as e:
          error("invalid ACL element ignored: %r: %s", ac_text, e)
          raise
        else:
          acl.append(ac)
    return acl

def xattrs_from_bytes(bs, offset=0):
  ''' Decode an XAttrs from some bytes, return the xattrs dictionary.
  '''
  xattrs = {}
  while offset < len(bs):
    name, offset = get_bss(bs, offset)
    data, offset = get_bsdata(bs, offset)
    if name in xattrs:
      warning("repeated name, ignored: %r", name)
    else:
      xattrs[name] = data
  return xattrs

# This is a direct dict subclass for memory efficiency.
class Meta(dict, Transcriber):
  ''' Inode metadata: times, permissions, ownership etc.

      This is a dictionary with the following keys:
      * `'u'`: owner
      * `'g'`: group owner
      * `'a'`: ACL
      * `'c'`: st_ctime, last change to inode/metadata
      * `'m'`: modification time, a float
      * `'su'`: setuid
      * `'sg'`: setgid
      * `'pathref'`: pathname component for symlinks
      * `'x'`: xattrs
  '''

  transcribe_prefix = 'M'

  def __init__(self, mapping=None):
    dict.__init__(self)
    self._lock = RLock()
    self._xattrs = {}
    dict.__setitem__(self, 'x', self._xattrs)
    self.acl = ACL()
    self._ctime = 0
    if mapping:
      for k, v in mapping.items():
        self[k] = v

  def __repr__(self):
    return ':'.join( (type(self).__name__, str(self)) )

  def textencode(self):
    ''' Return the encoding of this Meta as text.
    '''
    d = self._as_dict()
    if all(k in ('u', 'g', 'a', 'm', 'su', 'sg') for k in d.keys()):
      # these are all "safe" fields - use the compact encoding
      encoded = ';'.join(
          ':'.join((
              k,
              "%f" % v if isinstance(v, float) else str(v)
          ))
          for k, v in sorted(d.items())
      )
    else:
      # use the more verbose safe JSON encoding
      encoded = json.dumps(d, separators=(',', ':'))
    return encoded

  def _as_dict(self):
    ''' A dictionary usable for JSON or the compact transcription.
    '''
    d = dict(self)
    # drop xattrs if empty
    xa = d.get('x', {})
    if not xa and 'x' in d:
      del d['x']
    # convert ACL to str, leave other types alone
    d = dict(
        (k, (str(v) if isinstance(v, ACL) else v))
        for k, v in sorted(d.items())
    )
    return d

  def transcribe_inner(self, T, fp):
    ''' Transcribe the Meta.
    '''
    d = self._as_dict()
    return transcribe_mapping(d, fp, T=T)

  @classmethod
  def parse_inner(cls, T, s, offset, stopchar, prefix):
    ''' Parse a Meta transcription.
    '''
    if prefix != 'M':
      raise ValueError("expected prefix='M', got: %r" % (prefix,))
    m, offset = parse_mapping(s, offset, stopchar=stopchar, T=T)
    return cls(m), offset

  def __getitem__(self, k):
    if k == 'c':
      return self._ctime
    return dict.__getitem__(self, k)

  def __setitem__(self, k, v):
    # intercept some keys
    if k == 'c':
      v = float(v)
      self._ctime = v
      return
    if k == 'x':
      for xk, xv in v.items():
        self.setxattr(xk, xv)
      return
    if k == 'a':
      if isinstance(v, str):
        v = ACL.from_str(v)
      elif not isinstance(v, ACL):
        raise ValueError("not an ACL: %r" % (v,))
    elif k in ('m',):
      v = float(v)
    dict.__setitem__(self, k, v)
    self._ctime = time.time()

  def __delitem__(self, k):
    dict.__delitem__(self, k)
    self._ctime = time.time()

  @classmethod
  def from_text(cls, metatext):
    ''' Construct a new Meta from `metatext`.
    '''
    M = cls()
    M.update_from_text(metatext)
    return M

  def update_from_text(self, metatext):
    ''' Update the Meta fields from the supplied metatext.
    '''
    if metatext.startswith('{'):
      # wordy JSON encoding of metadata
      metadata = json.loads(metatext)
      kvs = metadata.items()
    else:
      # old style compact metadata
      kvs = []
      for metafield in metatext.split(';'):
        metafield = metafield.strip()
        if not metafield:
          continue
        try:
          k, v = metafield.split(':', 1)
        except ValueError:
          error("ignoring bad metatext field (no colon): %r", metafield)
          continue
        else:
          kvs.append((k, v))
    for k, v in kvs:
      if k == 'x':
        # update the xattrs from `v`, which should be a dict
        for xk, xv in v.items():
          self.setxattr(xk, xv)
      else:
        self[k] = v

  @property
  def user(self):
    ''' Return the username associated with this Meta's owner.
    '''
    u = self.get('u')
    if u is not None and not isinstance(u, str):
      u = username(u)
    return u

  @user.setter
  def user(self, u):
    ''' Set the owner (user) of this Meta.
    '''
    self['u'] = u

  @user.deleter
  def user(self):
    ''' Remove the owner of this Meta.
    '''
    if 'u' in self:
      del self['u']

  @property
  def uid(self):
    ''' Return the user id associated with this Meta's owner.
    '''
    u = self.get('u')
    if u is not None and not isinstance(u, int):
      u = userid(u)
    return u

  @uid.setter
  def uid(self, u):
    ''' Set the owner (user) of this Meta.
        Saves the user name of the supplied uid, or the uid if the
        user name cannot be looked up.
    '''
    user = username(u)
    if user is not None:
      u = user
    self['u'] = u

  @uid.deleter
  def uid(self):
    ''' Remove the owner (user) of this Meta.
    '''
    if 'u' in self:
      del self['u']

  @property
  def group(self):
    ''' Return the groupname associated with this Meta's group owner.
    '''
    g = self.get('g')
    if g is not None and not isinstance(g, str):
      g = username(g)
    return g

  @group.setter
  def group(self, g):
    ''' Set the group owner of this Meta.
    '''
    self['g'] = g

  @group.deleter
  def group(self):
    ''' Remove the group owner of this Meta.
    '''
    if 'g' in self:
      del self['g']

  @property
  def gid(self):
    ''' Return the group id associated with this Meta's group owner.
    '''
    g = self.get('g')
    if g is not None and not isinstance(g, int):
      g = groupid(g)
    return g

  @gid.setter
  def gid(self, g):
    ''' Set the owner (user) of this Meta.
        Saves the group name of the supplied gid, or the gid if the
        group name cannot be looked up.
    '''
    group = groupname(g)
    if group is not None:
      g = group
    self['g'] = g

  @gid.deleter
  def gid(self):
    ''' Remove the owner (user) of this Meta.
    '''
    if 'g' in self:
      del self['g']

  @property
  def mtime(self):
    ''' Return the time of last data modification, or 0.0.
    '''
    return self.get('m', 0.0)

  @mtime.setter
  def mtime(self, when):
    ''' Set the time of last modification.
    '''
    self['m'] = when

  @property
  @locked
  def acl(self):
    ''' Return the live list of AC instances, decoded at need.
    '''
    return self['a']

  @acl.setter
  @locked
  def acl(self, ac_L):
    ''' Rewrite the ACL with a list of AC instances.
    '''
    self['a'] = ac_L

  @property
  def setuid(self):
    ''' Return whether this Meta is setuid.
    '''
    return self.get('su', False)

  @setuid.setter
  def setuid(self, flag):
    ''' Set the setuidness of this Meta.
    '''
    if flag:
      self['su'] = flag
    elif 'su' in self:
      del self['su']

  @property
  def setgid(self):
    ''' Return whether this Meta is setgid.
    '''
    return self.get('sg', False)

  @setgid.setter
  def setgid(self, flag):
    ''' Set the setgidness of this Meta.
    '''
    if flag:
      self['sg'] = flag
    elif 'sg' in self:
      del self['sg']

  def chmod(self, mode):
    ''' Apply UNIX permissions to ACL.
    '''
    self.setuid = bool(mode & S_ISUID)
    self.setgid = bool(mode & S_ISGID)
    acl = ACL()
    acl.append(AC_Owner( *permbits_to_allow_deny( (mode>>6)&7 ) ))
    acl.append(AC_Group( *permbits_to_allow_deny( (mode>>3)&7 ) ))
    acl.append(AC_Other( *permbits_to_allow_deny( mode&7 ) ))
    acl.extend(ac for ac in self.acl if ac.audience not in ('o', 'g', '*'))
    self.acl = acl

  def update_from_stat(self, st):
    ''' Apply the contents of a stat object to this Meta.
    '''
    self.mtime = float(st.st_mtime)
    self.uid = st.st_uid
    self.gid = st.st_gid
    self.chmod(st.st_mode & 0o777)
    # TODO: setuid, setgid, sticky

  @property
  def unix_perm_bits(self):
    ''' Return unix-mode-bits.
        *Note*: excludes the type bits (S_IFREG etc).

        For ACLs with more than one user or group this is only an
        approximation.
    '''
    perms = 0
    for ac in self.acl:
      if ac.audience == 'o':
        perms |= ac.unixmode << 6
      elif ac.audience == 'g':
        perms |= ac.unixmode << 3
      elif ac.audience == '*':
        perms |= ac.unixmode
      else:
        warning("Meta.unix_perms: ignoring ACL element %s", ac)
    # TODO: sticky
    if self.setuid:
      perms |= S_ISUID
    if self.setgid:
      perms |= S_ISUID
    return perms

  @property
  def pathref(self):
    ''' Return the `pathref`, the path used by SymlinkDirents.
        Returns None if there is no pathref.
    '''
    return self.get('pathref')

  @pathref.setter
  def pathref(self, newref):
    ''' Set the `pathref` on a SymlinkDirent.
    '''
    self['pathref'] = newref

  def access(
      self, access_mode,
      access_uid=None, access_gid=None,
      default_uid=None, default_gid=None
  ):
    ''' POSIX like access call, accepting os.access `access_mode`.

        Parameters:
        * `access_mode`: a bitmask of os.{R_OK,W_OK,X_OK} as for
          the os.access function.
        * `access_uid`: the effective uid of the querying user.
        * `access_gid`: the effective gid of the querying user.
        * `default_uid`: the reference uid to use if this Meta.uid is None
        * `default_gid`: the reference gid to use if this Meta.gid is None

        If the Meta has no uid or `access_uid == Meta.uid`,
        use the owner permissions.
        Otherwise if the Meta has no gid or `access_gid == Meta.gid`,
        use the group permissions.
        Otherwise use the "other" permissions.
    '''
    X("META.ACCESS...")
    u = self.uid
    if u is None:
      u = default_uid
    g = self.gid
    if g is None:
      g = default_gid
    perms = self.unix_perm_bits
    if access_mode & os.R_OK:
      if u is None or ( access_uid is not None and access_uid == u):
        if not (perms>>6) & 4:
          return False
      elif g is None or ( access_gid is not None and access_gid == g):
        if not (perms>>3) & 4:
          return False
      elif not perms & 4:
        return False
    if access_mode & os.W_OK:
      if u is None or ( access_uid is not None and access_uid == u):
        if not (perms>>6) & 2:
          return False
      elif g is None or ( access_gid is not None and access_gid == g):
        if not (perms>>3) & 2:
          return False
      elif not perms & 2:
        return False
    if access_mode & os.X_OK:
      if u is None or ( access_uid is not None and access_uid == u):
        if not (perms>>6) & 1:
          return False
      elif g is None or ( access_gid is not None and access_gid == g):
        if not (perms>>3) & 1:
          return False
      elif not perms & 1:
        return False
    return True

  @staticmethod
  def xattrify(xkv):
    ''' Convert value `xkv` to str as though an ISO8859-1 encoding.

        We keep a mapping of str->str for the JSON encoding.
    '''
    if isinstance(xkv, (int, float)):
      xkv = str(xkv)
    if isinstance(xkv, str):
      xkv = xkv.encode('utf-8')
    if isinstance(xkv, bytes):
      return xkv.decode('iso8859-1')
    raise TypeError("cannot convert to bytes: %r" % (xkv,))

  def getxattr(self, xk, xv_default):
    ''' Return the bytes value for key `kx`, or `xv_default` if missing.
    '''
    k = self.xattrify(xk)
    try:
      v = self._xattrs[k]
    except KeyError:
      return xv_default
    # convert the str back into bytes
    return v.encode('iso8859-1')

  def setxattr(self, xk, xv):
    ''' Set the value for key `xk` to `xv`.
        Accepted types include: bytes, str, int, float.
    '''
    xk = self.xattrify(xk)
    xv = self.xattrify(xv)
    self._xattrs[xk] = xv
    dict.__setitem__(self, 'x', self._xattrs)

  def delxattr(self, xk):
    ''' Delete the key `xk` if present.
    '''
    xk = self.xattrify(xk)
    if xk in self._xattrs:
      del self._xattrs[xk]

  def listxattrs(self):
    ''' Return the xattr keys, a list of bytes.
    '''
    return [ xk.encode('iso8859-1') for xk in self._xattrs.keys() ]

  @property
  def mime_type(self):
    ''' Return the file's `user.mime_type` xattr, or None.
    '''
    return self.getxattr('user.mime_type', None)

  @mime_type.setter
  def mime_type(self, new_type):
    ''' Set the file's `user.mime_type` xattr.
    '''
    self.setxattr('user.mime_type', new_type)

  @mime_type.deleter
  def mime_type(self):
    ''' Delete the file's `user.mime_type` xattr.
    '''
    self.delxattr('user.mime_type')

register_transcriber(Meta)



#!/usr/bin/env python3
#

''' Conversion functions.
'''

from functools import partial
from os.path import expanduser, isabs as isabspath, join as joinpath
from cs.lex import skipwhite
from cs.pfx import Pfx
from cs.units import multiparse as multiparse_units, \
    BINARY_BYTES_SCALE, DECIMAL_BYTES_SCALE, DECIMAL_SCALE

def get_integer(s, offset):
  ''' Parse an integer followed by an optional scale and return computed value.
  '''
  return multiparse_units(
      s,
      (BINARY_BYTES_SCALE, DECIMAL_BYTES_SCALE, DECIMAL_SCALE),
      offset
  )

def scaled_value(s):
  ''' Convert a scaled value such as "8 GiB" into an int.
  '''
  value, offset = get_integer(s, 0)
  offset = skipwhite(s, offset)
  if offset < len(s):
    raise ValueError("unparsed text: %r" % (s[offset:],))
  return value

def convert_param(params, key, *, decoder=None):
  ''' Convert a parameter.
  '''
  param = params.get(key)
  if isinstance(param, str):
    with Pfx("%s: %r", key, param):
      params[key] = decoder(param)

def convert_param_int(params, key):
  ''' Convert an integer paramater to an int.
  '''
  return convert_param(params, key, decoder=int)

def convert_param_scaled_int(params, key):
  ''' Convert a scaled value into an int.
  '''
  return convert_param(params, key, decoder=scaled_value)

def expand_path(path, basedir=None):
  ''' Expand a path specification.
  '''
  path = expanduser(path)
  if basedir and not isabspath(path):
    path = joinpath(basedir, path)
  return path

def convert_param_path(params, key, basedir=None):
  ''' Convert a path parameter to an absolute pathname.
  '''
  return convert_param(params, key, decoder=partial(expand_path, basedir=basedir))



#!/usr/bin/python

''' Archive files.

    Archive files are records of data saved to a Store.
    Lines are appended to the archive file of the form:

      isodatetime unixtime dirent

    where unixtime is UNIX time (seconds since epoch) and dirent is the text
    transcription of a Dirent.
'''

from __future__ import print_function
from datetime import datetime
import errno
import os
from os.path import realpath, isfile
import time
from cs.fileutils import lockfile, shortpath
from cs.inttypes import Flags
from cs.lex import unctrl
from cs.logutils import warning, exception, debug
from cs.pfx import Pfx, gen as pfxgen
from cs.py.func import prop
from .dir import _Dirent
from .meta import NOUSERID, NOGROUPID

CopyModes = Flags('delete', 'do_mkdir', 'trust_size_mtime')

# shared mapping of archive paths to Archive instances
_ARCHIVES = {}

def Archive(path, mapping=None, missing_ok=False, weird_ok=False):
  ''' Return an Archive for the named file.
      Maintains a mapping of issued Archives in order to reuse that
      same Archive for a given path.
  '''
  global _ARCHIVES
  if not path.endswith('.vt'):
    if weird_ok:
      warning("unusual Archive path: %r", path)
    else:
      raise ValueError("invalid Archive path (should end in '.vt'): %r" % (path,))
  if not missing_ok and not isfile(path):
    raise ValueError("not a file: %r" % (path,))
  if mapping is None:
    mapping = _ARCHIVES
  path = realpath(path)
  A = mapping.get(path)
  if A is None:
    mapping[path] = A = _Archive(path)
  return A

class _Archive(object):
  ''' Manager for an archive.vt file.
  '''

  def __init__(self, arpath):
    ''' Initialise this Archive.
        `arpath`: path to file holding the archive records
    '''
    self.path = arpath
    self._last = None
    self._last_s = None
    self.notify_update = []

  def __str__(self):
    return "Archive(%s)" % (shortpath(self.path),)

  @prop
  def last(self):
    ''' Return the last (unixtime, Dirent) from the file, or (None, None).
    '''
    last_entry = self._last
    if last_entry is None:
      for entry in self:
        last_entry = entry
      if last_entry is None:
        return None, None
      self._last = last_entry
    return last_entry

  @pfxgen
  def __iter__(self):
    ''' Generator yielding (unixtime, Dirent) from the archive file.
    '''
    path = self.path
    with Pfx(path):
      try:
        with open(path) as fp:
          entries = self.parse(fp)
          for when, E in entries:
            if when is None and E is None:
              break
            yield when, E
      except OSError as e:
        if e.errno == errno.ENOENT:
          return
        raise

  def update(self, E, when=None, previous=None, force=False, source=None):
    ''' Save the supplied Dirent `E` with timestamp `when`.
        Return the Dirent transcription.

        Parameters:
        * `E`: the Dirent to save.
        * `when`: the POSIX timestamp for the save, default now.
        * `previous`: optional previous Dirent transcription; defaults
          to the latest Transcription from of the Archive
        * `force`: append an entry even if the previous entry has the
          same transcription as `previous`, default False
        * `source`: optional source indicator for the update, default None
    '''
    assert isinstance(E, _Dirent), "expected E<%s> to be a _Dirent" % (type(E),)
    etc = E.name
    if not force:
      # see if we should discard this update
      if previous is None:
        previous = self._last_s
      if previous is not None:
        # do not save if the previous transcription is unchanged
        Es = str(E)
        if Es == previous:
          return Es
    if when is None:
      when = time.time()
    path = self.path
    with lockfile(path):
      with open(path, "a") as fp:
        s = self.write(fp, E, when=when, etc=etc)
    self._last = when, E
    self._last_s = s
    for notify in self.notify_update:
      try:
        notify(E, when=when, source=source)
      except Exception as e:
        exception(
            "notify[%s](%s,when=%s,source=%s): %s",
            notify, E, when, source, e)
    return s

  @staticmethod
  def write(fp, E, when=None, etc=None):
    ''' Write a Dirent to an open archive file. Return the Dirent transcription.

        Archive lines have the form:

            isodatetime unixtime transcribe(dirent) dirent.name
    '''
    if when is None:
      when = time.time()
    # produce a local time with knowledge of its timezone offset
    dt = datetime.fromtimestamp(when).astimezone()
    assert dt.tzinfo is not None
    # precompute strings to avoid corrupting the archive file
    iso_s = dt.isoformat()
    when_s = str(when)
    if isinstance(E, str):
      Es = E
    else:
      Es = str(E)
    etc_s = None if etc is None else unctrl(etc)
    fp.write(iso_s)
    fp.write(' ')
    fp.write(when_s)
    fp.write(' ')
    fp.write(Es)
    if etc is not None:
      fp.write(' ')
      fp.write(etc_s)
    fp.write('\n')
    fp.flush()
    return Es

  @staticmethod
  @pfxgen
  def parse(fp, first_lineno=1):
    ''' Parse lines from an open archive file, yield `(when, E)`.
    '''
    for lineno, line in enumerate(fp, first_lineno):
      with Pfx(str(lineno)):
        if not line.endswith('\n'):
          raise ValueError("incomplete? no trailing newline")
        line = line.strip()
        if not line or line.startswith('#'):
          continue
        # allow optional trailing text, which will be the E.name part normally
        fields = line.split(None, 3)
        _, unixtime, dent = fields[:3]
        when = float(unixtime)
        E, offset = _Dirent.from_str(dent)
        if offset != len(dent):
          warning("unparsed dirent text: %r", dent[offset:])
        ##info("when=%s, E=%s", when, E)
        yield when, E

  @staticmethod
  def read(fp, first_lineno=1):
    ''' Read the next entry from an open archive file, return (when, E).
        Return (None, None) at EOF.
    '''
    for when, E in _Archive.parse(fp, first_lineno=first_lineno):
      return when, E
    return None, None

def apply_posix_stat(src_st, ospath):
  ''' Apply a stat object to the POSIX OS object at `ospath`.
  '''
  with Pfx("apply_posix_stat(%r)", ospath):
    path_st = os.stat(ospath)
    if src_st.st_uid == NOUSERID or src_st.st_uid == path_st.st_uid:
      uid = -1
    else:
      uid = src_st.st_uid
    if src_st.st_gid == NOGROUPID or src_st.st_gid == path_st.st_gid:
      gid = -1
    else:
      gid = src_st.st_gid
    if uid != -1 or gid != -1:
      with Pfx("chown(uid=%d,gid=%d)", uid, gid):
        debug(
            "chown(%r,%d,%d) from %d:%d",
            ospath, uid, gid, path_st.st_uid, path_st.st_gid)
        try:
          os.chown(ospath, uid, gid)
        except OSError as e:
          if e.errno == errno.EPERM:
            warning("%s", e)
          else:
            raise
    st_perms = path_st.st_mode & 0o7777
    mst_perms = src_st.st_mode & 0o7777
    if st_perms != mst_perms:
      with Pfx("chmod(0o%04o)", mst_perms):
        debug("chmod(%r,0o%04o) from 0o%04o", ospath, mst_perms, st_perms)
        os.chmod(ospath, mst_perms)
    mst_mtime = src_st.st_mtime
    if mst_mtime > 0:
      st_mtime = path_st.st_mtime
      if mst_mtime != st_mtime:
        with Pfx("chmod(0o%04o)", mst_perms):
          debug(
              "utime(%r,atime=%s,mtime=%s) from mtime=%s",
              ospath, path_st.st_atime, mst_mtime, st_mtime)
          os.utime(ospath, (path_st.st_atime, mst_mtime))
